import { logger } from '@libp2p/logger';
import { anySignal } from 'any-signal';
import errCode from 'err-code';
import drain from 'it-drain';
import defer from 'p-defer';
import PQueue from 'p-queue';
const log = logger('libp2p:delegated-content-routing');
const DEFAULT_TIMEOUT = 30e3; // 30 second default
const CONCURRENT_HTTP_REQUESTS = 4;
const CONCURRENT_HTTP_REFS_REQUESTS = 2;
export var EventTypes;
(function (EventTypes) {
    EventTypes[EventTypes["SENDING_QUERY"] = 0] = "SENDING_QUERY";
    EventTypes[EventTypes["PEER_RESPONSE"] = 1] = "PEER_RESPONSE";
    EventTypes[EventTypes["FINAL_PEER"] = 2] = "FINAL_PEER";
    EventTypes[EventTypes["QUERY_ERROR"] = 3] = "QUERY_ERROR";
    EventTypes[EventTypes["PROVIDER"] = 4] = "PROVIDER";
    EventTypes[EventTypes["VALUE"] = 5] = "VALUE";
    EventTypes[EventTypes["ADDING_PEER"] = 6] = "ADDING_PEER";
    EventTypes[EventTypes["DIALING_PEER"] = 7] = "DIALING_PEER";
})(EventTypes || (EventTypes = {}));
/**
 * The types of messages set/received during DHT queries
 */
export var MessageType;
(function (MessageType) {
    MessageType[MessageType["PUT_VALUE"] = 0] = "PUT_VALUE";
    MessageType[MessageType["GET_VALUE"] = 1] = "GET_VALUE";
    MessageType[MessageType["ADD_PROVIDER"] = 2] = "ADD_PROVIDER";
    MessageType[MessageType["GET_PROVIDERS"] = 3] = "GET_PROVIDERS";
    MessageType[MessageType["FIND_NODE"] = 4] = "FIND_NODE";
    MessageType[MessageType["PING"] = 5] = "PING";
})(MessageType || (MessageType = {}));
/**
 * An implementation of content routing, using a delegated peer
 */
class DelegatedContentRouting {
    client;
    httpQueue;
    httpQueueRefs;
    started;
    abortController;
    /**
     * Create a new DelegatedContentRouting instance
     */
    constructor(client) {
        if (client == null) {
            throw new Error('missing ipfs http client');
        }
        this.client = client;
        this.started = false;
        this.abortController = new AbortController();
        // limit concurrency to avoid request flood in web browser
        // https://github.com/libp2p/js-libp2p-delegated-content-routing/issues/12
        this.httpQueue = new PQueue({
            concurrency: CONCURRENT_HTTP_REQUESTS
        });
        // sometimes refs requests take long time, they need separate queue
        // to not suffocate regular business
        this.httpQueueRefs = new PQueue({
            concurrency: CONCURRENT_HTTP_REFS_REQUESTS
        });
        const { protocol, host, port } = client.getEndpointConfig();
        log(`enabled DelegatedContentRouting via ${protocol}://${host}:${port}`);
    }
    isStarted() {
        return this.started;
    }
    start() {
        this.started = true;
    }
    stop() {
        this.httpQueue.clear();
        this.httpQueueRefs.clear();
        this.abortController.abort();
        this.abortController = new AbortController();
        this.started = false;
    }
    /**
     * Search the dht for providers of the given CID.
     *
     * - call `findProviders` on the delegated node.
     */
    async *findProviders(key, options = {}) {
        log('findProviders starts: %c', key);
        options.timeout = options.timeout ?? DEFAULT_TIMEOUT;
        const signal = anySignal([this.abortController.signal, options.signal]);
        const onStart = defer();
        const onFinish = defer();
        void this.httpQueue.add(async () => {
            onStart.resolve();
            return onFinish.promise;
        });
        try {
            await onStart.promise;
            for await (const event of this.client.dht.findProvs(key, {
                ...options,
                signal
            })) {
                if (event.name === 'PROVIDER') {
                    yield* event.providers.map(prov => {
                        const peerInfo = {
                            id: prov.id,
                            protocols: [],
                            multiaddrs: prov.multiaddrs
                        };
                        return peerInfo;
                    });
                }
            }
        }
        catch (err) {
            log.error('findProviders errored:', err);
            throw err;
        }
        finally {
            signal.clear();
            onFinish.resolve();
            log('findProviders finished: %c', key);
        }
    }
    /**
     * Announce to the network that the delegated node can provide the given key.
     *
     * Currently this uses the following hack
     * - delegate is one of bootstrap nodes, so we are always connected to it
     * - call block stat on the delegated node, so it fetches the content
     * - call dht provide with the passed cid
     *
     * N.B. this must be called for every block in the dag you want provided otherwise
     * the delegate will only be able to supply the root block of the dag when asked
     * for the data by an interested peer.
     */
    async provide(key, options = {}) {
        log('provide starts: %c', key);
        options.timeout = options.timeout ?? DEFAULT_TIMEOUT;
        const signal = anySignal([this.abortController.signal, options.signal]);
        const onStart = defer();
        const onFinish = defer();
        void this.httpQueue.add(async () => {
            onStart.resolve();
            return onFinish.promise;
        });
        try {
            await onStart.promise;
            await this.client.block.stat(key, {
                ...options,
                signal
            });
            await drain(this.client.dht.provide(key, {
                ...options,
                signal
            }));
        }
        catch (err) {
            log.error('provide errored:', err);
            throw err;
        }
        finally {
            signal.clear();
            onFinish.resolve();
            log('provide finished: %c', key);
        }
    }
    /**
     * Stores a value in the backing key/value store of the delegated content router.
     * This may fail if the delegated node's content routing implementation does not
     * use a key/value store, or if the delegated operation fails.
     */
    async put(key, value, options = {}) {
        log('put value start: %b', key);
        options.timeout = options.timeout ?? DEFAULT_TIMEOUT;
        const signal = anySignal([this.abortController.signal, options.signal]);
        const onStart = defer();
        const onFinish = defer();
        void this.httpQueue.add(async () => {
            onStart.resolve();
            return onFinish.promise;
        });
        try {
            await onStart.promise;
            await drain(this.client.dht.put(key, value, {
                ...options,
                signal
            }));
        }
        catch (err) {
            log.error('put errored:', err);
            throw err;
        }
        finally {
            signal.clear();
            onFinish.resolve();
            log('put finished: %b', key);
        }
    }
    /**
     * Fetches an value from the backing key/value store of the delegated content router.
     * This may fail if the delegated node's content routing implementation does not
     * use a key/value store, or if the delegated operation fails.
     */
    async get(key, options = {}) {
        log('get value start: %b', key);
        options.timeout = options.timeout ?? DEFAULT_TIMEOUT;
        const signal = anySignal([this.abortController.signal, options.signal]);
        const onStart = defer();
        const onFinish = defer();
        void this.httpQueue.add(async () => {
            onStart.resolve();
            return onFinish.promise;
        });
        try {
            await onStart.promise;
            for await (const event of this.client.dht.get(key, {
                ...options,
                signal
            })) {
                if (event.name === 'VALUE') {
                    log('get value finished: %b', key);
                    return event.value;
                }
            }
            throw errCode(new Error('Not found'), 'ERR_NOT_FOUND');
        }
        catch (err) {
            log.error('put errored:', err);
            throw err;
        }
        finally {
            signal.clear();
            onFinish.resolve();
            log('put finished: %b', key);
        }
    }
}
export function delegatedContentRouting(client) {
    return () => new DelegatedContentRouting(client);
}
//# sourceMappingURL=index.js.map