import net from 'net';
import * as mafmt from '@multiformats/mafmt';
import { logger } from '@libp2p/logger';
import { toMultiaddrConnection } from './socket-to-conn.js';
import { TCPListener } from './listener.js';
import { multiaddrToNetConfig } from './utils.js';
import { AbortError, CodeError } from '@libp2p/interfaces/errors';
import { CODE_CIRCUIT, CODE_P2P, CODE_UNIX } from './constants.js';
import { symbol } from '@libp2p/interface-transport';
const log = logger('libp2p:tcp');
class TCP {
    constructor(components, options = {}) {
        this.opts = options;
        this.components = components;
        if (components.metrics != null) {
            this.metrics = {
                dialerEvents: components.metrics.registerCounterGroup('libp2p_tcp_dialer_events_total', {
                    label: 'event',
                    help: 'Total count of TCP dialer events by type'
                })
            };
        }
    }
    get [symbol]() {
        return true;
    }
    get [Symbol.toStringTag]() {
        return '@libp2p/tcp';
    }
    async dial(ma, options) {
        options.keepAlive = options.keepAlive ?? true;
        // options.signal destroys the socket before 'connect' event
        const socket = await this._connect(ma, options);
        // Avoid uncaught errors caused by unstable connections
        socket.on('error', err => {
            log('socket error', err);
        });
        const maConn = toMultiaddrConnection(socket, {
            remoteAddr: ma,
            socketInactivityTimeout: this.opts.outboundSocketInactivityTimeout,
            socketCloseTimeout: this.opts.socketCloseTimeout,
            metrics: this.metrics?.dialerEvents
        });
        const onAbort = () => {
            maConn.close().catch(err => {
                log.error('Error closing maConn after abort', err);
            });
        };
        options.signal?.addEventListener('abort', onAbort, { once: true });
        log('new outbound connection %s', maConn.remoteAddr);
        const conn = await options.upgrader.upgradeOutbound(maConn);
        log('outbound connection %s upgraded', maConn.remoteAddr);
        options.signal?.removeEventListener('abort', onAbort);
        if (options.signal?.aborted === true) {
            conn.close().catch(err => {
                log.error('Error closing conn after abort', err);
            });
            throw new AbortError();
        }
        return conn;
    }
    async _connect(ma, options) {
        if (options.signal?.aborted === true) {
            throw new AbortError();
        }
        return await new Promise((resolve, reject) => {
            const start = Date.now();
            const cOpts = multiaddrToNetConfig(ma);
            const cOptsStr = cOpts.path ?? `${cOpts.host ?? ''}:${cOpts.port}`;
            log('dialing %j', cOpts);
            const rawSocket = net.connect(cOpts);
            const onError = (err) => {
                err.message = `connection error ${cOptsStr}: ${err.message}`;
                this.metrics?.dialerEvents.increment({ error: true });
                done(err);
            };
            const onTimeout = () => {
                log('connection timeout %s', cOptsStr);
                this.metrics?.dialerEvents.increment({ timeout: true });
                const err = new CodeError(`connection timeout after ${Date.now() - start}ms`, 'ERR_CONNECT_TIMEOUT');
                // Note: this will result in onError() being called
                rawSocket.emit('error', err);
            };
            const onConnect = () => {
                log('connection opened %j', cOpts);
                this.metrics?.dialerEvents.increment({ connect: true });
                done();
            };
            const onAbort = () => {
                log('connection aborted %j', cOpts);
                this.metrics?.dialerEvents.increment({ abort: true });
                rawSocket.destroy();
                done(new AbortError());
            };
            const done = (err) => {
                rawSocket.removeListener('error', onError);
                rawSocket.removeListener('timeout', onTimeout);
                rawSocket.removeListener('connect', onConnect);
                if (options.signal != null) {
                    options.signal.removeEventListener('abort', onAbort);
                }
                if (err != null) {
                    reject(err);
                    return;
                }
                resolve(rawSocket);
            };
            rawSocket.on('error', onError);
            rawSocket.on('timeout', onTimeout);
            rawSocket.on('connect', onConnect);
            if (options.signal != null) {
                options.signal.addEventListener('abort', onAbort);
            }
        });
    }
    /**
     * Creates a TCP listener. The provided `handler` function will be called
     * anytime a new incoming Connection has been successfully upgraded via
     * `upgrader.upgradeInbound`.
     */
    createListener(options) {
        return new TCPListener({
            ...options,
            maxConnections: this.opts.maxConnections,
            backlog: this.opts.backlog,
            closeServerOnMaxConnections: this.opts.closeServerOnMaxConnections,
            socketInactivityTimeout: this.opts.inboundSocketInactivityTimeout,
            socketCloseTimeout: this.opts.socketCloseTimeout,
            metrics: this.components.metrics
        });
    }
    /**
     * Takes a list of `Multiaddr`s and returns only valid TCP addresses
     */
    filter(multiaddrs) {
        multiaddrs = Array.isArray(multiaddrs) ? multiaddrs : [multiaddrs];
        return multiaddrs.filter(ma => {
            if (ma.protoCodes().includes(CODE_CIRCUIT)) {
                return false;
            }
            if (ma.protoCodes().includes(CODE_UNIX)) {
                return true;
            }
            return mafmt.TCP.matches(ma.decapsulateCode(CODE_P2P));
        });
    }
}
export function tcp(init = {}) {
    return (components = {}) => {
        return new TCP(components, init);
    };
}
//# sourceMappingURL=index.js.map