import { connect } from 'it-ws/client';
import { multiaddrToUri as toUri } from '@multiformats/multiaddr-to-uri';
import { AbortError } from '@libp2p/interfaces/errors';
import pDefer from 'p-defer';
import { logger } from '@libp2p/logger';
import { isBrowser, isWebWorker } from 'wherearewe';
import { createListener } from './listener.js';
import { socketToMaConn } from './socket-to-conn.js';
import * as filters from './filters.js';
import { symbol } from '@libp2p/interface-transport';
const log = logger('libp2p:websockets');
class WebSockets {
    constructor(init) {
        this.init = init;
    }
    get [Symbol.toStringTag]() {
        return '@libp2p/websockets';
    }
    get [symbol]() {
        return true;
    }
    async dial(ma, options) {
        log('dialing %s', ma);
        options = options ?? {};
        const socket = await this._connect(ma, options);
        const maConn = socketToMaConn(socket, ma);
        log('new outbound connection %s', maConn.remoteAddr);
        const conn = await options.upgrader.upgradeOutbound(maConn);
        log('outbound connection %s upgraded', maConn.remoteAddr);
        return conn;
    }
    async _connect(ma, options) {
        if (options?.signal?.aborted === true) {
            throw new AbortError();
        }
        const cOpts = ma.toOptions();
        log('dialing %s:%s', cOpts.host, cOpts.port);
        const errorPromise = pDefer();
        const errfn = (err) => {
            log.error('connection error:', err);
            errorPromise.reject(err);
        };
        const rawSocket = connect(toUri(ma), this.init);
        if (rawSocket.socket.on != null) {
            rawSocket.socket.on('error', errfn);
        }
        else {
            rawSocket.socket.onerror = errfn;
        }
        if (options.signal == null) {
            await Promise.race([rawSocket.connected(), errorPromise.promise]);
            log('connected %s', ma);
            return rawSocket;
        }
        // Allow abort via signal during connect
        let onAbort;
        const abort = new Promise((resolve, reject) => {
            onAbort = () => {
                reject(new AbortError());
                rawSocket.close().catch(err => {
                    log.error('error closing raw socket', err);
                });
            };
            // Already aborted?
            if (options?.signal?.aborted === true) {
                onAbort();
                return;
            }
            options?.signal?.addEventListener('abort', onAbort);
        });
        try {
            await Promise.race([abort, errorPromise.promise, rawSocket.connected()]);
        }
        finally {
            if (onAbort != null) {
                options?.signal?.removeEventListener('abort', onAbort);
            }
        }
        log('connected %s', ma);
        return rawSocket;
    }
    /**
     * Creates a Websockets listener. The provided `handler` function will be called
     * anytime a new incoming Connection has been successfully upgraded via
     * `upgrader.upgradeInbound`
     */
    createListener(options) {
        return createListener({ ...this.init, ...options });
    }
    /**
     * Takes a list of `Multiaddr`s and returns only valid Websockets addresses.
     * By default, in a browser environment only DNS+WSS multiaddr is accepted,
     * while in a Node.js environment DNS+{WS, WSS} multiaddrs are accepted.
     */
    filter(multiaddrs) {
        multiaddrs = Array.isArray(multiaddrs) ? multiaddrs : [multiaddrs];
        if (this.init?.filter != null) {
            return this.init?.filter(multiaddrs);
        }
        // Browser
        if (isBrowser || isWebWorker) {
            return filters.wss(multiaddrs);
        }
        return filters.all(multiaddrs);
    }
}
export function webSockets(init = {}) {
    return () => {
        return new WebSockets(init);
    };
}
//# sourceMappingURL=index.js.map