export class Network {
    /**
     * @param {import('@libp2p/interface-libp2p').Libp2p} libp2p
     * @param {import('./bitswap').Bitswap} bitswap
     * @param {import('./stats').Stats} stats
     * @param {object} [options]
     * @param {boolean} [options.b100Only]
     * @param {MultihashHasherLoader} [options.hashLoader]
     * @param {number} [options.maxInboundStreams=32]
     * @param {number} [options.maxOutboundStreams=32]
     * @param {number} [options.incomingStreamTimeout=30000]
     */
    constructor(libp2p: import('@libp2p/interface-libp2p').Libp2p, bitswap: import('./bitswap').Bitswap, stats: import('./stats').Stats, options?: {
        b100Only?: boolean | undefined;
        hashLoader?: import("./types").MultihashHasherLoader | undefined;
        maxInboundStreams?: number | undefined;
        maxOutboundStreams?: number | undefined;
        incomingStreamTimeout?: number | undefined;
    } | undefined);
    _log: import("@libp2p/logger").Logger;
    _libp2p: import("@libp2p/interface-libp2p").Libp2p;
    _bitswap: import("./bitswap").Bitswap;
    _protocols: string[];
    _stats: import("./stats").Stats;
    _running: boolean;
    /**
     * @private
     * @param {PeerId} peerId
     */
    private _onPeerConnect;
    /**
     * @private
     * @param {PeerId} peerId
     */
    private _onPeerDisconnect;
    /**
     * Handles both types of incoming bitswap messages
     *
     * @private
     * @param {object} connection
     * @param {Stream} connection.stream - A duplex iterable stream
     * @param {Connection} connection.connection - A libp2p Connection
     */
    private _onConnection;
    _hashLoader: import("./types").MultihashHasherLoader | undefined;
    _maxInboundStreams: number;
    _maxOutboundStreams: number;
    _incomingStreamTimeout: number;
    start(): Promise<void>;
    /** @type {string[]} */
    _registrarIds: string[] | undefined;
    stop(): Promise<void>;
    /**
     * Find providers given a `cid`.
     *
     * @param {CID} cid
     * @param {object} [options]
     * @param {AbortSignal} [options.signal]
     * @returns {AsyncIterable<Provider>}
     */
    findProviders(cid: CID, options?: {
        signal?: AbortSignal | undefined;
    } | undefined): AsyncIterable<Provider>;
    /**
     * Find the providers of a given `cid` and connect to them.
     *
     * @param {CID} cid
     * @param {object} [options]
     * @param {AbortSignal} [options.signal]
     */
    findAndConnect(cid: CID, options?: {
        signal?: AbortSignal | undefined;
    } | undefined): Promise<void>;
    /**
     * Tell the network we can provide content for the passed CID
     *
     * @param {CID} cid
     * @param {object} [options]
     * @param {AbortSignal} [options.signal]
     */
    provide(cid: CID, options?: {
        signal?: AbortSignal | undefined;
    } | undefined): Promise<void>;
    /**
     * Connect to the given peer
     * Send the given msg (instance of Message) to the given peer
     *
     * @param {PeerId} peer
     * @param {Message} msg
     */
    sendMessage(peer: PeerId, msg: Message): Promise<void>;
    /**
     * Connects to another peer
     *
     * @param {PeerId|Multiaddr} peer
     * @param {object} [options]
     * @param {AbortSignal} [options.signal]
     * @returns {Promise<Connection>}
     */
    connectTo(peer: PeerId | Multiaddr, options?: {
        signal?: AbortSignal | undefined;
    } | undefined): Promise<Connection>;
    /**
     * @private
     * @param {PeerId} peer
     * @param {Map<string, Uint8Array>} blocks
     */
    private _updateSentStats;
}
export type PeerId = import('@libp2p/interface-peer-id').PeerId;
export type CID = import('multiformats').CID;
export type Multiaddr = import('@multiformats/multiaddr').Multiaddr;
export type Connection = import('@libp2p/interface-connection').Connection;
export type Stream = import('@libp2p/interface-connection').Stream;
export type MultihashHasherLoader = import('./types').MultihashHasherLoader;
export type Provider = {
    id: PeerId;
    multiaddrs: Multiaddr[];
};
export type Duplex = import('it-stream-types').Duplex<Uint8Array>;
import { BitswapMessage as Message } from "./message/index.js";
//# sourceMappingURL=network.d.ts.map